"""E2e test cases that encode and decodes a trashcan file."""
import unittest

import io
import os.path as path
from pathlib import Path
import struct

import numpy as np

import pycodec2

TEST_DATA_DIR = path.join(path.dirname(path.realpath(__file__)), 'data')

class TrashcanTestCase(unittest.TestCase):
    def test_encodes_and_decodes_trashcan(self):
        c2 = pycodec2.Codec2(1200)
        with io.BytesIO() as output:
            with open(path.join(TEST_DATA_DIR, 'trashcan.raw'), 'rb') as input:
                INT16_BYTE_SIZE = 2
                PACKET_SIZE = c2.samples_per_frame() * INT16_BYTE_SIZE
                STRUCT_FORMAT = '{}h'.format(c2.samples_per_frame())
                while True:
                    packet = input.read(PACKET_SIZE)
                    if len(packet) != PACKET_SIZE:
                        break
                    packet = np.array(struct.unpack(STRUCT_FORMAT, packet),
                                      dtype=np.int16)
                    encoded = c2.encode(packet)
                    packet = c2.decode(encoded)
                    output.write(struct.pack(STRUCT_FORMAT, *packet))

            output.seek(0)
            with open(path.join(TEST_DATA_DIR, 'trashcan-1200-decoded.raw'), 'rb') as expected_output:
                self.assertEqual(output.read(), expected_output.read())



# Synced with
# https://github.com/drowe67/codec2/commits/25014a290752fb830390ff24a612a9694319a47a/src/codec2.h
cdef extern from 'codec2/codec2.h':
  cdef enum:
    CODEC2_MODE_3200 = 0
    CODEC2_MODE_2400 = 1
    CODEC2_MODE_1600 = 2
    CODEC2_MODE_1400 = 3
    CODEC2_MODE_1300 = 4
    CODEC2_MODE_1200 = 5
    CODEC2_MODE_700C = 8
    CODEC2_MODE_450  = 10
    CODEC2_MODE_450PWB  = 11
  cdef struct CODEC2:
    pass

  CODEC2* codec2_create(int mode)
  void codec2_destroy(CODEC2* codec2_state)
  void codec2_encode(CODEC2 *codec2_state,
      unsigned char *bytes,
      short speech_in[])
  void codec2_decode(CODEC2 *codec2_state,
      short speech_out[],
      const unsigned char *bytes)
  void codec2_decode_ber(CODEC2 *codec2_state,
      short speech_out[],
      const unsigned char *bytes,
      float ber_est)
  int codec2_samples_per_frame(CODEC2 *codec2_state)
  int codec2_bits_per_frame(CODEC2 *codec2_state)
  int codec2_bytes_per_frame(CODEC2 *codec2_state);
  void codec2_set_lpc_post_filter(CODEC2 *codec2_state,
      int enable,
      int bass_boost,
      float beta,
      float gamma)
  int codec2_get_spare_bit_index(CODEC2 *codec2_state)
  int codec2_rebuild_spare_bit(CODEC2 *codec2_state, char unpacked_bits[])
  void codec2_set_natural_or_gray(CODEC2 *codec2_state, int gray)
  void codec2_set_softdec(CODEC2 *c2, float *softdec);
  float codec2_get_energy(CODEC2 *codec2_state, const unsigned char *bits);

  # support for ML and VQ experiments
  void codec2_open_mlfeat(CODEC2 *codec2_state, char *feat_filename, char *model_filename);
  void codec2_load_codebook(CODEC2 *codec2_state, int num, char *filename);
  float codec2_get_var(CODEC2 *codec2_state);
  float *codec2_enable_user_ratek(CODEC2 *codec2_state, int *K);

  # 700C post filter and equaliser
  void codec2_700c_post_filter(CODEC2 *codec2_state, int en);
  void codec2_700c_eq(CODEC2 *codec2_state, int en);



#cython: language_level=3
# The language level specification is needed. I don't have it in my build file,
# so I use it per-module.
from codec2 cimport *

import math

import numpy as np
cimport numpy as cnp
ctypedef cnp.int8_t CHAR_DTYPE_t
ctypedef cnp.int16_t SHORT_DTYPE_t
ctypedef cnp.int_t INT_DTYPE_t

_modes = {
    450  : CODEC2_MODE_450,
    451  : CODEC2_MODE_450PWB,
    700  : CODEC2_MODE_700C,
    1200 : CODEC2_MODE_1200,
    1300 : CODEC2_MODE_1300,
    1400 : CODEC2_MODE_1400,
    1600 : CODEC2_MODE_1600,
    2400 : CODEC2_MODE_2400,
    3200 : CODEC2_MODE_3200,
    }

cdef class Codec2:
  '''Wrapper for codec2 state and its functions.

  Initialization method expects an integer defining expected bitrate per
  second.'''
  cdef CODEC2 *_c_codec2_state

  def __cinit__(self, mode):
    self._c_codec2_state = codec2_create(_modes[mode])
    if self._c_codec2_state is NULL:
      raise MemoryError()

  def __dealloc__(self):
    if self._c_codec2_state is not NULL:
      codec2_destroy(self._c_codec2_state)

  def encode(self, cnp.ndarray[SHORT_DTYPE_t, ndim=1] speech_in):
    '''Encode samples with codec2.

    Encode the given ndarray of samples to bits represented as a byte array.'''
    assert len(speech_in) % self.samples_per_frame() == 0
    frames = len(speech_in) // self.samples_per_frame()
    bit_count = frames * self.bits_per_frame()
    bits = b'\x00' * int(math.ceil(bit_count / 8.0))
    codec2_encode(self._c_codec2_state, bits, <short*>cnp.PyArray_DATA(speech_in))
    return bits

  def decode(self, bytes frames):
    '''Decode a byte array into an ndarray of samples'''
    assert len(frames) >= self.bytes_per_frame()
    cdef cnp.ndarray[SHORT_DTYPE_t, ndim=1] speech_out
    frame_count = int(math.floor(len(frames) / self.bytes_per_frame()))
    sample_count = frame_count * self.samples_per_frame()
    speech_out = np.empty(sample_count, dtype=np.int16, order='C')
    codec2_decode(self._c_codec2_state, <short *>(cnp.PyArray_DATA(speech_out)), frames)
    return speech_out

  def decode_ber(self, bytes frames, float ber_est):
    assert len(frames) >= self.bytes_per_frame()
    cdef cnp.ndarray[SHORT_DTYPE_t, ndim=1] speech_out
    frame_count = len(frames) // self.bytes_per_frame()
    sample_count = frame_count * self.samples_per_frame()
    speech_out = np.empty(sample_count, dtype=np.int16, order='C')
    codec2_decode_ber(self._c_codec2_state,
        <short *>(cnp.PyArray_DATA(speech_out)),
        frames,
        ber_est)
    return speech_out

  def samples_per_frame(self):
    return codec2_samples_per_frame(self._c_codec2_state)

  def bits_per_frame(self):
    return codec2_bits_per_frame(self._c_codec2_state)

  def bytes_per_frame(self):
    return codec2_bytes_per_frame(self._c_codec2_state)

  def set_lpc_post_filter(self,
      int enable,
      int bass_boost,
      float beta,
      float gamma):
    codec2_set_lpc_post_filter(self._c_codec2_state,
        enable,
        bass_boost,
        beta,
        gamma)

  def get_spare_bit_index(self):
    return codec2_get_spare_bit_index(self._c_codec2_state)

  def rebuild_spare_bit(self, cnp.ndarray[char, ndim=1] unpacked_bits):
    return codec2_rebuild_spare_bit(self._c_codec2_state, <char *>cnp.PyArray_DATA(unpacked_bits))

  def set_natural_or_gray(self, int gray):
    codec2_set_natural_or_gray(self._c_codec2_state, gray)

  def set_softdec(self, float softdec):
    codec2_set_softdec(self._c_codec2_state, &softdec)

  def get_energy(self,  cnp.ndarray[unsigned char, ndim=1] bits):
      return codec2_get_energy(self._c_codec2_state, <unsigned char*>(cnp.PyArray_DATA(bits)))

  # support for ML and VQ experiments
  def open_mlfeat(self, bytearray feat_filename, bytearray model_filename):
      codec2_open_mlfeat(self._c_codec2_state, feat_filename, model_filename)

  def load_codebook(self, bytearray feat_filename, int num, bytearray filename):
      codec2_load_codebook(self._c_codec2_state, num, filename)

  def get_var(self):
      return codec2_get_var(self._c_codec2_state)

  def enable_user_ratek(self):
      cdef int K
      cdef float *user_rate_k = codec2_enable_user_ratek(self._c_codec2_state, &K)
      return (user_rate_k[0], K)

  # 700C post filter and equaliser
  def post_filter_700c(self, int en):
      codec2_700c_post_filter(self._c_codec2_state, en)

  def eq_700c(self, int en):
      codec2_700c_eq(self._c_codec2_state, en)



import Cython.Build
from Cython.Build import cythonize
import numpy as np
from setuptools import Extension, setup

VERSION = '2.1.0'

ext_modules = [
    Extension("pycodec2",
              [
                  "pycodec2/codec2.pxd",
                  "pycodec2/pycodec2.pyx",
              ],
              include_dirs=[np.get_include()],
              # This line guarantees that we do not numpy API deprecated in
              # 1.23.
              define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_23_API_VERSION")],
              libraries=["codec2"]) # Unix-like specific
]


setup(
  name = "pycodec2",
  packages = ['pycodec2'],
  version = VERSION,
  description = 'A Cython wrapper for codec2',
  long_description = open('README.md').read(),
  long_description_content_type = 'text/markdown',
  author = 'Grzegorz Milka',
  author_email = 'grzegorzmilka@gmail.com',
  url = 'https://github.com/gregorias/pycodec2',
  keywords = ['codec2', 'audio', 'voice'],
  classifiers = ['Topic :: Multimedia :: Sound/Audio :: Speech',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 3',
                 'Development Status :: 5 - Production/Stable',
                 ],
  ext_modules = cythonize(ext_modules),
  cmdclass={'build_ext': Cython.Build.build_ext},
)
